#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include "minigame.h"

static int st_stone_width;
static int st_stone_height;
static HBITMAP st_stone_bitmap;

static int st_foodchopper_width;
static int st_foodchopper_height;
static HBITMAP st_foodchopper_bitmap;

static int st_ball_width;
static int st_ball_height;
static HBITMAP st_ball_bitmap;

#define SCISSORS_ANIMATION_COUNT (3)
#define SCISSORS_ANIMATION_TIMER (10)
static int st_scissors_width[SCISSORS_ANIMATION_COUNT];
static int st_scissors_height[SCISSORS_ANIMATION_COUNT];
static HBITMAP st_scissors_bitmap[SCISSORS_ANIMATION_COUNT];

FallingObject *g_falling_object_top;
FallingObject *g_falling_object_tail;

static void move_stone(FallingObject *obj) {
  obj->y += 4;
}

static void move_foodchopper(FallingObject *obj) {
  obj->u.foodchopper.real_y += obj->u.foodchopper.velocity;
  obj->y = (int)obj->u.foodchopper.real_y;
  obj->u.foodchopper.velocity += 1.0;
}

static void move_ball(FallingObject *obj) {
  obj->u.ball.real_x += obj->u.ball.x_vector;
  obj->x = (int)obj->u.ball.real_x;
  obj->y += BALL_SPEED;
}

static void move_scissors(FallingObject *obj) {
  int image_idx;

  image_idx = obj->u.scissors.animation_counter / SCISSORS_ANIMATION_TIMER;
  obj->x = obj->u.scissors.center_x + st_scissors_width[image_idx] / 2;
  obj->width = st_scissors_width[image_idx];
  obj->height = st_scissors_height[image_idx];
  obj->bitmap = st_scissors_bitmap[image_idx];
  obj->y += 3;

  if (obj->u.scissors.now_openning) {
    obj->u.scissors.animation_counter++;
    if (obj->u.scissors.animation_counter
	>= SCISSORS_ANIMATION_COUNT * SCISSORS_ANIMATION_TIMER) {
      obj->u.scissors.now_openning = 0;
    }
  } else {
    obj->u.scissors.animation_counter--;
    if (obj->u.scissors.animation_counter < 0) {
      obj->u.scissors.now_openning = 1;
    }
  }
}

static void (*st_move_func_table[])(FallingObject *obj) = {
  move_stone,
  move_foodchopper,
  move_ball,
  move_scissors
};

static void get_bitmap_rect(HBITMAP hbitmap, int *width_p, int *height_p) {
  BITMAP bitmap;

  GetObject(hbitmap, sizeof(BITMAP), &bitmap);
  *width_p = bitmap.bmWidth;
  *height_p = bitmap.bmHeight;
}

void initialize_falling_objects(void) {
  int i;
  char buf[1024];

  /*  */
  st_stone_bitmap = LoadImage(NULL, "stone.bmp", IMAGE_BITMAP, 0, 0,
			      LR_LOADFROMFILE);
  get_bitmap_rect(st_stone_bitmap, &st_stone_width, &st_stone_height);

  /*  */
  st_foodchopper_bitmap = LoadImage(NULL, "foodchopper.bmp", IMAGE_BITMAP, 0, 0,
				    LR_LOADFROMFILE);
  get_bitmap_rect(st_foodchopper_bitmap,
		  &st_foodchopper_width, &st_foodchopper_height);

  /* {[ */
  st_ball_bitmap = LoadImage(NULL, "ball.bmp", IMAGE_BITMAP, 0, 0,
			     LR_LOADFROMFILE);
  get_bitmap_rect(st_ball_bitmap, &st_ball_width, &st_ball_height);

  /* nT~ */
  for (i = 0; i < SCISSORS_ANIMATION_COUNT; i++) {
    sprintf(buf, "scissors%d.bmp", i);
    st_scissors_bitmap[i] = LoadImage(NULL, buf, IMAGE_BITMAP, 0, 0,
				      LR_LOADFROMFILE);
    get_bitmap_rect(st_scissors_bitmap[i],
		    &st_scissors_width[i], &st_scissors_height[i]);
  }

  g_falling_object_top = NULL;
  g_falling_object_tail = NULL;
}

void draw_falling_objects(HWND hwnd) {
  FallingObject *pos;

  for (pos = g_falling_object_top; pos != NULL; pos = pos->next) {
    fill_black_rectangle(hwnd, pos->prev_x, pos->prev_y,
			 pos->prev_width, pos->prev_height);
    draw_bitmap(hwnd, pos->bitmap, pos->x, pos->y, pos->width, pos->height);
    pos->prev_x = pos->x;
    pos->prev_y = pos->y;
    pos->prev_width = pos->width;
    pos->prev_height = pos->height;
  }
}

static void create_new_object() {
  FallingObject *new_p;
  int rand_val;

  new_p = malloc(sizeof(FallingObject));

  new_p->y = 0;

  rand_val = rand();
  if (rand_val < 8000) {
    new_p->kind = FOODCHOPPER_FALLING_OBJECT;
    new_p->width = st_foodchopper_width;
    new_p->height = st_foodchopper_height;
    new_p->bitmap = st_foodchopper_bitmap;
    new_p->x = rand() % (g_window_width - new_p->width);
    new_p->u.foodchopper.real_y = 0;
    new_p->u.foodchopper.velocity = 0;
  }else if (rand_val < 12000) {
    new_p->kind = BALL_FALLING_OBJECT;
    new_p->width = st_ball_width;
    new_p->height = st_ball_height;
    new_p->bitmap = st_ball_bitmap;
    new_p->x = rand() % (g_window_width - new_p->width);
    {
      new_p->u.ball.x_vector
        = ((double)(g_player_x + g_player_width / 2
                    - (new_p->x + st_ball_width / 2))
	   / ((double)g_window_height / BALL_SPEED));
    }
    new_p->u.ball.real_x = new_p->x;
  } else if (rand_val < 18000) {
    int image_idx;

    new_p->kind = SCISSORS_FALLING_OBJECT;
    new_p->u.scissors.animation_counter
      = rand() % (SCISSORS_ANIMATION_COUNT * SCISSORS_ANIMATION_TIMER);
    image_idx = new_p->u.scissors.animation_counter / SCISSORS_ANIMATION_TIMER;
    new_p->width = st_scissors_width[image_idx];
    new_p->height = st_scissors_height[image_idx];
    new_p->bitmap = st_scissors_bitmap[image_idx];
    new_p->u.scissors.center_x = rand() % g_window_width;
    new_p->x = new_p->u.scissors.center_x - new_p->width / 2;
    new_p->u.scissors.now_openning = 1;
  } else {
    new_p->kind = STONE_FALLING_OBJECT;
    new_p->width = st_stone_width;
    new_p->height = st_stone_height;
    new_p->bitmap = st_stone_bitmap;
    new_p->x = rand() % (g_window_width - new_p->width);
  }
  new_p->prev_x = new_p->x;
  new_p->prev_y = new_p->y;
  new_p->prev_width = new_p->width;
  new_p->prev_height = new_p->height;

  new_p->next = NULL;

  if (g_falling_object_top == NULL) {
    g_falling_object_top = new_p;
    g_falling_object_tail = new_p;
  } else {
    g_falling_object_tail->next = new_p;
    g_falling_object_tail = new_p;
  }
}

void move_falling_objects(void) {
  FallingObject *pos;
  FallingObject *prev;
  FallingObject *temp;

  prev = NULL;
  for (pos = g_falling_object_top; pos != NULL; ) {
    if (pos->y > g_window_height) {
      if (prev == NULL) {
        g_falling_object_top = pos->next;
      } else {
	prev->next = pos->next;
      }
      if (pos->next == NULL) {
        g_falling_object_tail = prev;
      }
      temp = pos->next;
      free(pos);
      pos = temp;
    } else {
      st_move_func_table[pos->kind](pos);

      prev = pos;
      pos = pos->next;
    }
  }
  if (rand() < 2000) {
    create_new_object();
  }
}

int check_crash(int x, int y, int width, int height) {
  FallingObject *pos;

  for (pos = g_falling_object_top; pos != NULL; pos = pos->next) {
    if (pos->x + pos->width > x && pos->x < x + width
        && pos->y + pos->height > y && pos->y < y + height) {
      return 1;
    }
  }
  return 0;
}

