#include <windows.h>
#include <stdio.h>
#include <time.h>
#include "minigame.h"

#define MESSAGE_LOOP_EXIT (-1)

int g_window_width;
int g_window_height;

int g_player_width;
int g_player_height;
int g_player_x;
int g_player_y;
static int st_prev_player_x;
static HBITMAP st_player_l_bitmap;
static HBITMAP st_player_r_bitmap;
static HBITMAP st_player_dead_bitmap;

typedef enum {
  LEFT_DIRECTION,
  RIGHT_DIRECTION,
} PlayerDirection;
static PlayerDirection st_player_direction;

LRESULT CALLBACK window_proc(HWND hwnd, UINT msg, WPARAM wp, LPARAM lp) {
  switch (msg) {
  case WM_DESTROY:
    PostQuitMessage(0);
    return 0;
  case WM_TIMER:
    return MESSAGE_LOOP_EXIT;
  }
  return DefWindowProc(hwnd, msg, wp, lp);
}

static void initialize(HWND hwnd) {
  RECT rect;
  BITMAP bitmap;

  GetClientRect(hwnd, &rect);
  g_window_width = rect.right;
  g_window_height = rect.bottom;

  st_player_l_bitmap = LoadImage(NULL, "player_l.bmp", IMAGE_BITMAP, 0, 0,
				 LR_LOADFROMFILE);
  st_player_r_bitmap = LoadImage(NULL, "player_r.bmp", IMAGE_BITMAP, 0, 0,
				 LR_LOADFROMFILE);
  st_player_dead_bitmap = LoadImage(NULL, "player_dead0.bmp",
                                    IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE);
  GetObject(st_player_l_bitmap, sizeof(BITMAP), &bitmap);
  g_player_width = bitmap.bmWidth;
  g_player_height = bitmap.bmHeight;

  g_player_y = g_window_height - g_player_height - 1;
  g_player_x = (g_window_width - g_player_width) / 2;
  st_prev_player_x = g_player_x;
  st_player_direction = LEFT_DIRECTION;
}

static void timed_message_loop(HWND hwnd, int millisecond)
{
  MSG msg;
  int gm_ret;
  UINT timer;

  timer = SetTimer(hwnd, 1, millisecond, NULL);
  for (;;) {
    gm_ret = GetMessage(&msg, NULL, 0, 0);
    if (gm_ret == 0 || gm_ret == -1) {
      exit(0);
    }
    if (DispatchMessage(&msg) == MESSAGE_LOOP_EXIT) {
      break;
    }
  }
  KillTimer(hwnd, 1);
}

void draw_bitmap(HWND hwnd, HBITMAP hbmp,
		 int x, int y, int width, int height)
{
  HBITMAP hbmp_prev;
  HDC hdc;
  HDC hdc_bmp;

  hdc = GetDC(hwnd);
  hdc_bmp = CreateCompatibleDC(hdc);
  hbmp_prev = SelectObject(hdc_bmp, hbmp);
  BitBlt(hdc, x, y, width, height, hdc_bmp, 0, 0, SRCCOPY);
  SelectObject(hdc_bmp, hbmp_prev);
  DeleteDC(hdc_bmp);
  ReleaseDC(hwnd, hdc);
}

void
fill_black_rectangle(HWND hwnd, int x, int y, int width, int height) {
  HDC hdc;
  
  hdc = GetDC(hwnd);
  SelectObject(hdc, GetStockObject(BLACK_BRUSH));
  SelectObject(hdc, GetStockObject(NULL_PEN));
  Rectangle(hdc, x, y, x + width + 1, y + height + 1);
  ReleaseDC(hwnd, hdc);
}

static void draw_player(HWND hwnd) {
  HBITMAP player_bitmap;

  fill_black_rectangle(hwnd, st_prev_player_x, g_player_y,
                       g_player_width, g_player_height);
  if (st_player_direction == LEFT_DIRECTION) {
    player_bitmap = st_player_l_bitmap;
  } else {
    player_bitmap = st_player_r_bitmap;
  }
  draw_bitmap(hwnd, player_bitmap, g_player_x, g_player_y,
              g_player_width, g_player_height);
  st_prev_player_x = g_player_x;
}

static void move_player(void)
{
  if (GetKeyState(VK_LEFT) < 0) {
    if (g_player_x - 5 >= 0) {
      g_player_x -= 5;
    }
    st_player_direction = LEFT_DIRECTION;
  }
  if (GetKeyState(VK_RIGHT) < 0) {
    if (g_player_x + g_player_width < g_window_width) {
      g_player_x += 5;
    }
    st_player_direction = RIGHT_DIRECTION;
  }
}

static void main_loop(HWND hwnd) {
  for (;;) {
    draw_player(hwnd);
    draw_falling_objects(hwnd);
    if (check_crash(g_player_x, g_player_y,
                    g_player_width, g_player_height)) {
      break;
    }
    move_player();
    move_falling_objects();
    timed_message_loop(hwnd, 20);
  }
}

static void gameover_loop(HWND hwnd) {
  for (;;) {
    draw_bitmap(hwnd, st_player_dead_bitmap, g_player_x, g_player_y,
                g_player_width, g_player_height);
    draw_falling_objects(hwnd);
    move_falling_objects();
    timed_message_loop(hwnd, 20);
  }
}

int WINAPI WinMain(HINSTANCE hInstance , HINSTANCE hPrevInstance,
                   PSTR lpCmdLine , int nCmdShow ) {
  HWND hwnd;
  WNDCLASS winc;

  winc.style = CS_HREDRAW | CS_VREDRAW;
  winc.lpfnWndProc = window_proc;
  winc.cbClsExtra = 0;
  winc.cbWndExtra = 0;
  winc.hInstance = hInstance;
  winc.hIcon = LoadIcon(NULL, IDI_APPLICATION);
  winc.hCursor = LoadCursor(NULL, IDC_ARROW);
  winc.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
  winc.lpszMenuName = NULL;
  winc.lpszClassName = TEXT("MiniGame");

  if (!RegisterClass(&winc)) {
    return -1;
  }

  hwnd = CreateWindow(TEXT("MiniGame") , TEXT("Mini GAME") ,
                      WS_OVERLAPPEDWINDOW | WS_VISIBLE,
                      0, 0, 400, 600,
                      NULL, NULL, hInstance, NULL);
  if (hwnd== NULL) {
    return -1;
  }

  srand((unsigned int)time(NULL));
  initialize(hwnd);
  initialize_falling_objects();
  main_loop(hwnd);
  gameover_loop(hwnd);

  return 0;
}